import readlineSync from "readline-sync";
import fs from "fs";
import { AIAgent } from "./commands/main.js";
import colors from "colors/safe.js";
import { RecordingTypes } from "./commands/command.js";
import { GetPageStatus } from "./commands/get_page_status.js";
import { TaskProgress } from "./tasks/task_progress.js";

let aiAgent = null;
let context = null;
let run = true;
while (run) {
  var projectPath = readlineSync.question("Project home folder (for ../cucumber_demo press Enter)?\n");
  if (!projectPath) {
    projectPath = "../cucumber_demo";
  }
  if (!fs.existsSync(projectPath)) {
    console.log("Folder does not exist.");
    continue;
  }
  aiAgent = new AIAgent(projectPath);
  context = {};
  await aiAgent.initProjectAndEnvironment(context);
  console.log("found pages: " + colors.green(aiAgent.project.getPagesNames()));
  while (run) {
    let commands = ["Page operations", "Close project"];
    let functions = [pageOperators, closeProject];
    let cont = await commandsSelect(commands, functions);
    if (!cont) {
      break;
    }
  }
}

async function commandsSelect(commands, functions) {
  while (run) {
    let index = readlineSync.keyInSelect(commands, "What do you want to do?");
    if (index < 0) {
      return false;
    }
    await functions[index]();
    //    return continueSelection;
  }
}
async function pageOperators() {
  let commands = [
    "Navigate to page",
    "Navigate by instructions",
    "Create new page",
    "Create new step",
    "Test step",
    "Rerun task",
    "Generate code for task",
    "Dump page",
  ];
  let functions = [
    navigateToPage,
    navigateByInstractions,
    createNewPage,
    createNewStep,
    testStep,
    rerunTask,
    generateCodeForTask,
    dump,
  ];
  return await commandsSelect(commands, functions);
}
async function closeProject() {
  await aiAgent.closeSession(context);
  return true;
}
async function navigateToPage() {
  let pages = aiAgent.project.getPagesNames();
  let index = readlineSync.keyInSelect(pages, "Select page?");
  if (index < 0) {
    return false;
  }
  await closeBrowserIfOpen();
  const taskProgress = new TaskProgress(-2);
  await aiAgent.loginAndNavigate(context, pages[index], taskProgress);
  return false;
}
async function navigateByInstractions() {
  var task = readlineSync.question("Provide instractions:\n");
  if (!task) {
    return false;
  }
  closeBrowserIfOpen();
  await aiAgent.mainRecordingLoop(context, task, null, null, RecordingTypes.NAVIGATION);
  return true;
}
async function createNewPage() {
  var name = readlineSync.question("Page name?\n");
  if (!name) {
    return false;
  }
  var path = readlineSync.question("Page path?\n");
  if (!path) {
    return false;
  }
  var login_required =
    readlineSync.keyInSelect([true, false], "Login required?", {
      cancel: false,
    }) === 0;

  var description = readlineSync.question("Page description?\n");
  if (!description) {
    return false;
  }
  await aiAgent.project.createNewPage(name, path, login_required, description, false);
  return true;
}
async function dump() {
  const pageStatus = new GetPageStatus();
  const result = await pageStatus.run(context);
  console.log(result.pageText);
}

async function createNewStep() {
  let pages = aiAgent.project.getPagesNames();
  let index = readlineSync.keyInSelect(pages, "Select page?");
  if (index < 0) {
    return false;
  }
  let pageName = pages[index];
  let task = readlineSync.question("The task to perform:\n");
  if (!task) {
    return false;
  }
  let userDataInput = readlineSync.question("User data (json)?\n");
  let userData = null;
  if (!userDataInput) {
    userData = {};
  } else {
    userData = JSON.parse(userDataInput);
  }
  let options = ["Given", "When", "Then"];
  let stepType = readlineSync.keyInSelect(options, "Step type?", {
    cancel: false,
  });
  let step = options[stepType];
  await closeBrowserIfOpen();
  let taskProgress = new TaskProgress();
  await aiAgent.createNewStep(context, pageName, task, userData, step, taskProgress);
  return true;
}
async function rerunTask() {
  let taskId = readlineSync.question("Provide a task id:\n");
  if (!taskId) {
    return false;
  }
  let taskProgress = new TaskProgress(parseInt(taskId), true);
  await closeBrowserIfOpen();
  await aiAgent.createNewStep(
    context,
    taskProgress.taskParameters.pageName,
    taskProgress.taskParameters.task,
    taskProgress.taskParameters.userData,
    taskProgress.taskParameters.stepType,
    new TaskProgress()
  );
  return true;
}
async function generateCodeForTask() {
  let taskId = readlineSync.question("Provide a task id:\n");
  if (!taskId) {
    return false;
  }
  let taskProgress = new TaskProgress(parseInt(taskId), true);
  taskProgress.shadow = true;
  await closeBrowserIfOpen();
  await aiAgent.createNewStep(
    context,
    taskProgress.taskParameters.pageName,
    taskProgress.taskParameters.task,
    taskProgress.taskParameters.userData,
    taskProgress.taskParameters.stepType,
    taskProgress,
    true
  );
  return true;
}
async function closeBrowserIfOpen() {
  await aiAgent.closeSession(context, false);
}
async function testStep() {
  let pages = aiAgent.project.getPagesNames();
  let index = readlineSync.keyInSelect(pages, "Select page?");
  if (index < 0) {
    return false;
  }
  let pageName = pages[index];
  let page = aiAgent.project.getPage(pageName);
  if (!page) {
    throw new Error(`Page ${pageName} not found`);
  }
  let methods = page.getMethodsNames();
  index = readlineSync.keyInSelect(methods, "Select method?");
  if (index < 0) {
    return false;
  }
  let methodName = methods[index];
  await closeBrowserIfOpen();
  const taskProgress = new TaskProgress(-2);
  await aiAgent.testStep(context, pageName, methodName, taskProgress);
  return true;
}
