import { Command, Types, postCommandSleep } from "./command.js";
import logger from "../../logger.js";
import { AccessibilityDump } from "../tasks/accessibility_dump.js";
import { UiActionError } from "./ui_action_error.js";

class ClickElement extends Command {
  constructor(id, recording) {
    super(Types.CLICK, "click on element in the page", [{ name: "id", description: "element id number" }], recording);
    this.id = id;
  }
  // eslint-disable-next-line
  async run(context, executeStep = null, taskProgress) {
    this.preRun(context, taskProgress);
    taskProgress.updateLastCommand({ aiInstruction: "click on element id: " + this.id });
    let element = null;
    if (executeStep) {
      element = executeStep.element;
    } else {
      for (let i = 0; i < context.elements.length; i++) {
        if (context.elements[i].localId === this.id) {
          element = context.elements[i];
        }
      }
      this.recordingStep.element = element;
    }

    if (!element) {
      logger.error(`Element with id ${this.id} not found`);
      taskProgress.updateLastCommand({ success: false });
      taskProgress.updateLastCommand({ error: `element with id: ${this.id} couldn't be identify` });
      throw new Error(`Element with id ${this.id} not found`);
    }
    // retry to click on the element
    let maxTry = 3;
    for (let i = 0; i < maxTry; i++) {
      logger.info(`Clicking on element: ${element.getReadableType()}: ${element.getText()}`);
      try {
        const options = this.getCommandOptions(taskProgress);
        let info = await context.stable.click(element.getUniqueLocatorObject(), options);
        taskProgress.updateLastCommand({ info: info });
        taskProgress.updateLastCommand({ success: true });
        logger.debug("clicked", JSON.stringify(info));
        break;
      } catch (err) {
        logger.error(`Failed to click on element: ${err}`);
        await new Promise((resolve) => setTimeout(resolve, 1000));
        if (i === maxTry - 1) {
          const error = new UiActionError(
            `Failed to click on element ${element.getReadableType()}: ${element.getText()}`,
            err
          );
          error.info = err.info;

          taskProgress.updateLastCommand({ info: error.info });
          taskProgress.updateLastCommand({ success: false });

          throw error;
        }
      }
    }

    context.stable.waitForPageLoad();
    await new Promise((resolve) => setTimeout(resolve, postCommandSleep + 1500));

    const task2 = new AccessibilityDump();
    await task2.execute(context);

    return this.postCommand(
      `clicked on element: ${element.getReadableType()}: ${element.getText()}`,
      taskProgress,
      context
    );
  }
}
export { ClickElement };
