import { getContext } from "automation_model";
import fs from "fs";
const postCommandSleep = 500;
import { logger } from "../../logger.js";
const Types = {
  CLICK: "click_element",
  NAVIGATE: "navigate",
  FILL: "fill_element",
  EXECUTE: "execute_page_method",
  OPEN: "open_environment",
  COMPLETE: "task_complete",
  ASK: "information_needed",
  GET_PAGE_STATUS: "get_page_status",
  CLICK_ROW_ACTION: "click_row_action",
  VERIFY_ELEMENT_EXIST: "verify_element_exist",
};
const Statuses = {
  PASSED: "passed",
  FAILED: "failed",
  UNKNOWN: "unknown",
};
const RecordingTypes = {
  NAVIGATION: "navigation",
  FORM: "form",
};
class Command {
  constructor(name, description, parameters = [], recording = null) {
    this.name = name;
    this.description = description;
    this.parameters = parameters;
    this.recording = recording;
    this.recordingStep = new RecordedStep(null, this.name, null);
    this.status = Statuses.UNKNOWN;
    this.guide = null;
    if (this.recording) {
      this.recording.addStep(this.recordingStep);
    }
  }
  // eslint-disable-next-line
  async run(context, executeStep = null, taskProgress = null) {
    throw new Error("Not implemented");
  }
  async preRun(context, taskProgress = null) {
    if (!context.environment) {
      throw new Error("Environment not found in context");
    }
    if (!context.browser) {
      const bcontext = await getContext(context.environment, false, logger);
      if (context.stable) {
        context.stable.logger = logger;
      }
      Object.assign(context, bcontext);
    }
    taskProgress.addProgressCommand();
  }
  postCommand = (longTermMemory, taskProgress, context) => {
    const result = new CommandResult();
    result.longTermMemory = longTermMemory;
    taskProgress.updateLastCommand({ memory: result.longTermMemory });
    taskProgress.updateLastCommand({ success: true });
    result.lastCommandResult = context.pageText;
    result.pageText = context.pageText;
    this.recordingStep.status = Statuses.PASSED;
    return result;
  };
  getCommandOptions = (taskProgress) => {
    const options = { screenshot: false };
    if (taskProgress.isActive()) {
      options.screenshot = true;
      options.screenshotPath = taskProgress.getScreenShotPath();
      taskProgress.updateLastCommand({ screenshotId: taskProgress.screenshotId });
    }
    return options;
  };

  getTextDescription = () => {
    let parametersText = "no args";
    if (this.parameters && this.parameters.length > 0) {
      // join parameters
      const paramsTexts = [];
      this.parameters.forEach((parameter) => {
        paramsTexts.push(`"${parameter.name}": "<${parameter.description.replaceAll(" ", "_")}>"`);
      });
      parametersText = paramsTexts.join(", ");
    }
    return this.name + ": " + this.description + ", args: " + parametersText;
  };
  getInJsonFormat = () => {
    let _guide = "";
    if (this.guide) {
      _guide = `  guide: "${this.guide}",\n`;
    }
    let json = `{ name: "${this.name}",\n  description: "${this.description}",\n${_guide}  parameters: [\n`;
    if (this.parameters && this.parameters.length > 0) {
      // join parameters
      for (let i = 0; i < this.parameters.length; i++) {
        const parameter = this.parameters[i];
        json += `    { name: "${parameter.name}", description: "${parameter.description}" },\n`;
      }
    }

    json += "  ]\n},";
    return json;
  };
}

class CommandResult {
  constructor() {
    this.longTermMemory = "";
    this.lastCommandResult = null;
    this.pageText = null;
  }
}
class RecordedStep {
  element;
  type;
  parameters;
  status;

  constructor(element, type, parameters) {
    this.element = element;
    this.type = type;
    this.parameters = parameters;
    this.dataSource = null;
    this.dataKey = null;
  }
}
class Recording {
  steps = [];
  locators = {};
  type;
  fileName;
  recordingsPath = "./recordings";
  imageNextIndex = 0;
  constructor(type) {
    this.type = type;
    this.fileName = `recording_${new Date().toJSON().replaceAll(":", "_")}.json`;
  }
  addStep(step) {
    this.steps.push(step);
  }
  getParametersMap() {
    const result = {};
    this.steps.forEach((step) => {
      if (step.type === Types.FILL) {
        result[step.dataKey] = step.parameters[0];
      }
    });
    return result;
  }
  saveToFile() {
    if (this.steps.length === 0) {
      return;
    }
    if (!fs.existsSync(this.recordingsPath)) {
      fs.mkdir(this.recordingsPath, { recursive: true });
    }
    fs.writeFileSync(this.recordingsPath + "/" + this.fileName, JSON.stringify(this));
  }
  loadFromFile(filePath) {
    const data = fs.readFileSync(filePath, "utf8");
    const recording = JSON.parse(data);

    Object.assign(this, recording);
  }
  // async saveScreenshot(page) {
  //   const screenshotPath = this.recordingsPath + "/screenshots";
  //   if (!fs.existsSync(screenshotPath)) {
  //     fs.mkdirSync(screenshotPath, { recursive: true });
  //   }
  //   await page.screenshot({ path: screenshotPath + `/${this.imageNextIndex}.png` });
  //   if (this.steps && this.steps.length > 0) {
  //     this.steps[this.steps.length - 1].screenshot = this.imageNextIndex;
  //   }
  //   this.imageNextIndex++;
  // }
}
export { Command, CommandResult, Types, Recording, RecordedStep, Statuses, RecordingTypes, postCommandSleep };
