import { Command, Types, postCommandSleep } from "./command.js";

import logger from "../../logger.js";
import { AccessibilityDump } from "../tasks/accessibility_dump.js";
import { UiActionError } from "./ui_action_error.js";

class FillElement extends Command {
  constructor(id, text, recording) {
    super(
      Types.FILL,
      "fill input element with text",
      [
        { name: "id", description: "element id number" },
        { name: "text", description: "the text to send to the input field" },
      ],
      recording
    );
    this.id = id;
    this.text = text;
    this.recordingStep.parameters = [this.text];
  }
  async run(context, executeStep = null, taskProgress) {
    this.preRun(context, taskProgress);
    taskProgress.updateLastCommand({ aiInstruction: "fill element id: " + this.id + " with text: " + this.text });
    let element = null;
    if (executeStep) {
      this.text = executeStep.parameters[0];
      element = executeStep.element;
    } else {
      for (let i = 0; i < context.elements.length; i++) {
        if (context.elements[i].localId === this.id) {
          element = context.elements[i];
        }
      }
      this.recordingStep.element = element;
      let found = false;
      if (context.userData) {
        const keys = Object.keys(context.userData);
        keys.forEach((key) => {
          if (context.userData[key] === this.text) {
            found = true;
            this.recordingStep.dataSource = "userData";
            this.recordingStep.dataKey = key;
          }
        });
      }
      if (context.environment && !found) {
        const keys = Object.keys(context.environment);
        keys.forEach((key) => {
          if (context.environment[key] === this.text) {
            this.recordingStep.dataSource = "environment";
            this.recordingStep.dataKey = key;
          }
        });
      }
    }

    if (!element) {
      taskProgress.updateLastCommand({ success: false });
      taskProgress.updateLastCommand({ error: `element with id: ${this.id} couldn't be identify` });
      throw new Error(`Element with id ${this.id} not found`);
    }
    logger.info(`Filled element: ${element.getReadableType()}: ${element.getText()} with text: ${this.text}`);
    try {
      const options = this.getCommandOptions(taskProgress);
      let info = await context.stable.fill(element.getUniqueLocatorObject(), this.text, options);
      taskProgress.updateLastCommand({ info: info });
      taskProgress.updateLastCommand({ success: true });
      logger.debug("clicked", JSON.stringify(info));
    } catch (err) {
      logger.error(`Failed to fill element: ${err}`);
      const error = new UiActionError(
        `Failed to fill element: ${element.getReadableType()}: ${element.getText()} with text: ${this.text}`,
        err
      );
      error.info = err.info;
      taskProgress.updateLastCommand({ info: error.info });
      taskProgress.updateLastCommand({ success: false });
      throw error;
    }
    context.stable.waitForPageLoad();
    await new Promise((resolve) => setTimeout(resolve, postCommandSleep + 1500));

    const task2 = new AccessibilityDump();
    await task2.execute(context);

    return this.postCommand(
      `filled element ${element.getReadableType()}: ${element.getText()} with text: ${this.text}`,
      taskProgress,
      context
    );
  }
}
export { FillElement };
