import { OpenEnvironment } from "./open_environment.js";
import { Project } from "../commands/project.js";
// import { fileURLToPath } from "url";
// import { dirname } from "path";
import { getCommandFromName, getCommandsSummary } from "./commands.js";
import { Recording, RecordingTypes, Types } from "./command.js";
import { generateCode } from "./playwright_codeget.js";
import logger from "../../logger.js";
import {
  isComplete,
  navigationNextCommand,
  getCucumberLine,
  getMethodName,
  getDummyData,
  isTaskCompleteSuccessfully,
} from "../tasks/openai/ai_ask.js";
import { GetPageStatus } from "./get_page_status.js";
import { ExecutePageMethod } from "./execute_page_method.js";
import { closeBrowser } from "automation_model";
import { Memory } from "./memory.js";
import { TaskProgress, TasksTypes } from "../tasks/task_progress.js";

import { UiActionError } from "./ui_action_error.js";
import { isRecordingFailed, generateRecoveryTasks } from "./ai_failure_identifier.js";

//const __filename = fileURLToPath(import.meta.url);
//const __dirname = dirname(__filename);

class AIAgent {
  longTermMemory = [];
  nextCommand = null;
  project = null;
  lastCommandResult = null;
  constructor(projectFolder) {
    this.projectFolder = projectFolder;
  }
  async initProjectAndEnvironment(context) {
    this.project = new Project(this.projectFolder);
    await this.project.load();
    context.environment = this.project.environment;
    context.project = this.project;
  }
  async createNewStep(
    context,
    pageName,
    task,
    userData = null,
    stepType = "Given",
    taskProgress = null,
    useRecording = false,
    login = true,
    previousTask = null
  ) {
    let recording = null;
    if (!useRecording) {
      taskProgress.setLoggerFile();
      taskProgress.startTask(TasksTypes.CREATE_STEP, task, "Login/Navigate", {
        pageName,
        task,
        userData,
        stepType,
      });

      logger.info(`Starting create new step: ${pageName} ${task}`);
      try {
        if (login) {
          await this.loginAndNavigate(context, pageName, taskProgress);
        }
      } catch (e) {
        taskProgress.setError(e);
        throw e;
      }
      taskProgress.updateProgress("Record steps");
      try {
        recording = await this.mainRecordingLoop(
          context,
          task,
          previousTask ? previousTask : "login and navigate to page: " + pageName,
          userData,
          RecordingTypes.FORM,
          taskProgress
        );
      } catch (e) {
        taskProgress.setError(e);
        throw e;
      }
      taskProgress.updateProgress("Generate code");
    } else {
      recording = new Recording(RecordingTypes.FORM);
      recording.loadFromFile(taskProgress.taskPath + "/rec.json");
    }
    try {
      const page = this.project.getPage(pageName);
      const generateCodeResult = generateCode(recording, page);
      page.insertElements(generateCodeResult.elements);
      console.log(page.fileContent);
      const description = task;
      const methodName = await getMethodName(description);
      page.addInfraCommand(methodName, description, generateCodeResult.parameters, generateCodeResult.codeLines);
      const newCucumberLine = await getCucumberLine(description, recording.getParametersMap(), stepType);
      page.addCucumberStep(
        stepType,
        newCucumberLine.substring((stepType + " ").length),
        methodName,
        generateCodeResult.parameters
      );

      page.save();
    } catch (e) {
      taskProgress.setError(e);
      taskProgress.updateProgress(null, false);
      throw e;
    }
    taskProgress.success = true;
    taskProgress.setDone("Done", true);
  }
  async testStep(context, pageName, methodName, taskProgress) {
    logger.info(`Starting test step: ${pageName} ${methodName}`);
    await this.loginAndNavigate(context, pageName, taskProgress);
    const page = this.project.getPage(pageName);
    if (!page) {
      throw new Error(`Page ${pageName} not found`);
    }
    const paramsTxt = page.getMethodParameters(methodName).join(",");
    let parameters = {};
    if (paramsTxt) {
      parameters = await getDummyData(paramsTxt);
    }
    const executePageMethod = new ExecutePageMethod(pageName, methodName, parameters);
    await executePageMethod.run(context);
  }
  async loginAndNavigate(context, pageName, taskProgress) {
    logger.info(`Starting login and navigate to page: ${pageName}`);
    await this.initProjectAndEnvironment(context);
    const page = this.project.getPage(pageName);
    if (!page) {
      throw new Error(`Page ${pageName} not found`);
    }

    let openEnvironment = new OpenEnvironment();
    await openEnvironment.run(context, null, taskProgress);

    if (page.isLoginRequired()) {
      let executePageMethod = new ExecutePageMethod(null, "login");
      await executePageMethod.run(context);
    }
    if (page.isNavigateOnInit()) {
      let executePageMethod = new ExecutePageMethod(pageName, "navigate");
      await executePageMethod.run(context);
    }
  }
  async runRecording(context, executeRecording) {
    logger.info("Starting recording execution");
    await this.initProjectAndEnvironment(context);
    const taskProgress = new TaskProgress(-2);
    taskProgress.shadow = true;
    for (let i = 0; i < executeRecording.steps.length; i++) {
      const step = executeRecording.steps[i];
      let command = {
        name: step.type,
        args: step.parameters,
      };
      this.nextCommand = getCommandFromName(command, null);
      const result = await this.nextCommand.run(context, step, taskProgress);
      this.longTermMemory.push(result.longTermMemory);
      this.lastCommandResult = result.lastCommandResult;
    }
    this.nextCommand = null;
  }

  /**
   * Will perform AI loop to record UI actions
   * @param {object} context
   * @param {string} task
   * @param {string} previousTask
   * @param {object} userData
   * @returns {Recording}
   */
  async mainRecordingLoop(
    context,
    task,
    previousTask,
    userData = null,
    type = RecordingTypes.FORM,
    taskProgress = null
  ) {
    logger.info("Starting main loop, task: " + task + ", type: " + type);
    await this.initProjectAndEnvironment(context);
    const recording = new Recording(type);
    if (taskProgress) {
      recording.recordingsPath = taskProgress.taskPath;
      recording.fileName = "rec.json";
    }

    context.userData = userData;
    const longMemory = new Memory();
    this.longTermMemory = [];
    if (!previousTask) {
      previousTask = context.task;
    }

    let commandObject = null;
    let memory = null;
    // eslint-disable-next-line
    while (true) {
      if (this.nextCommand === null) {
        if (type === RecordingTypes.FORM) {
          this.nextCommand = new GetPageStatus(null, previousTask);
        } else if (type === RecordingTypes.NAVIGATION) {
          this.nextCommand = new OpenEnvironment(recording);
        }
      }

      longMemory.add(this.nextCommand.name);

      if (commandObject && commandObject.thoughts) {
        longMemory.addLast(commandObject.thoughts);
      }
      if (taskProgress) {
        taskProgress.memory = longMemory.history;
        taskProgress.updateProgress();
      }
      try {
        let result = await this.nextCommand.run(context, null, taskProgress);
        recording.saveToFile();
        if (this.nextCommand.name === Types.COMPLETE) {
          taskProgress.memory = longMemory.history;
          taskProgress.updateProgress();
          this.nextCommand = null;
          return recording;
        }
        this.lastCommandResult = result.lastCommandResult;
        longMemory.addLast({ result: result.longTermMemory });
        longMemory.addLast({ status: "success" });
        if (taskProgress) {
          taskProgress.memory = longMemory.history;
          taskProgress.updateProgress();
        }
        if (isRecordingFailed(taskProgress)) {
          logger.error("Recording failed");
          generateRecoveryTasks(taskProgress);
        }
        //memory = this._createMemory(speak, result.longTermMemory);
      } catch (e) {
        let uiError = false;
        if (e instanceof UiActionError) {
          uiError = true;
          logger.error(e.message);
          logger.error(e.originalError, e.originalError.stack, e.originalError.message);
        } else {
          logger.error(e, e.stack, e.message);
        }
        let complete = await isComplete(task, memory);
        longMemory.addLast({
          status: uiError ? "failed - element identification" : "failed",
        });
        if (complete) {
          logger.info("early complete");
          taskProgress.memory = longMemory.history;
          taskProgress.updateProgress();
          this.nextCommand = null;
          return recording;
        }
        if (taskProgress) {
          taskProgress.memory = longMemory.history;
          taskProgress.result = false;
          taskProgress.error = e;
          taskProgress.updateProgress();
        }
        this.nextCommand = null;
        throw e;
        //memory = this._createMemory("", `Failed to execute: ${text}`);
      }
      commandObject = await navigationNextCommand(
        task,
        task,
        this.lastCommandResult,
        longMemory.toString(),
        JSON.stringify(context.environment),
        getCommandsSummary(this.project, type),
        userData ? JSON.stringify(userData) : "N/A"
      );

      this.nextCommand = getCommandFromName(commandObject.command, recording);
      if (this.nextCommand.name === Types.COMPLETE) {
        let result = await isTaskCompleteSuccessfully(task, task, this.lastCommandResult, longMemory.toString());
        if (!result.thoughts.success) {
          longMemory.add("task is not complete");
          longMemory.addLast({ result: result.thoughts.reasoning });
          longMemory.addLast({ status: "incomplete" });
          commandObject = await navigationNextCommand(
            task,
            task,
            this.lastCommandResult,
            longMemory.toString(),
            JSON.stringify(context.environment),
            getCommandsSummary(this.project, type),
            userData ? JSON.stringify(userData) : "N/A"
          );

          this.nextCommand = getCommandFromName(commandObject.command, recording);
        }
      }
    }
  }
  async closeSession(context, cleanProperties = true) {
    logger.info("Closing session");
    if (context.browser) {
      await closeBrowser(context.browser);
      context.browser = null;
      context.page = null;
      if (cleanProperties) {
        for (const property in context) {
          if (Object.prototype.hasOwnProperty.call(context, property)) {
            delete context[property];
          }
        }
      }
    }
  }
}

export { AIAgent };
