import { Command, Types } from "./command.js";
import logger from "../../logger.js";
import { AccessibilityDump } from "../tasks/accessibility_dump.js";
import { UiActionError } from "./ui_action_error.js";

class VerifyElementExist extends Command {
  constructor(role, name, recording) {
    super(
      Types.VERIFY_ELEMENT_EXIST,
      "verify element exist in the page",
      [
        { name: "role", description: "the role of the elmment" },
        { name: "name", description: "the name of the elmment" },
      ],
      recording
    );
    this.guide =
      "Use this function only if the task includes instructions to perform verification. Verification can only be done to elements that currently exists in the page.";
    this.role = role;
    this.elementName = name;
    this.recordingStep.parameters = [this.role, this.elementName];
  }
  async run(context, executeStep = null, taskProgress) {
    this.preRun(context, taskProgress);
    taskProgress.updateLastCommand({
      aiInstruction: `verify element ${this.role}: ${this.elementName} exist in the page`,
    });

    if (executeStep) {
      this.role = executeStep.parameters[0];
      this.elementName = executeStep.parameters[1];
    }

    // retry to click on the element
    const elementDescription = `${this.role} ${this.elementName}`;
    logger.info(`Verify element ${elementDescription} exist in the page`);
    try {
      const options = this.getCommandOptions(taskProgress);
      let info = await context.stable.verifyElementExistInPage(
        {
          role: [this.role, { name: this.elementName }],
        },
        options
      );

      taskProgress.updateLastCommand({ info });
      taskProgress.updateLastCommand({ success: true });
    } catch (err) {
      logger.error(`Element was not found: ${elementDescription}, error: ${err}`);
      await new Promise((resolve) => setTimeout(resolve, 1000));
      const error = new UiActionError(`Failed to find element ${elementDescription}`, err);
      error.info = err.info;

      taskProgress.updateLastCommand({ info: error.info });
      taskProgress.updateLastCommand({ success: false });

      throw error;
    }

    const task2 = new AccessibilityDump();
    await task2.execute(context);

    return this.postCommand(`element ${elementDescription} exist in the page`, taskProgress, context);
  }
}
export { VerifyElementExist };
