import { AIAgent } from "./commands/main.js";
import logger from "../logger.js";
import fs from "fs";
import path from "path";
import { findNextIdInFolder, TaskProgress } from "./tasks/task_progress.js";
// args: projectDir, feature_file, step_definition_file
const SupportedTypes = ["Given", "When", "Then", "And", "But"];

class Step {
  constructor(line) {
    this.type = null;
    if (!line) {
      throw new Error("line is required");
    }
    this.line = line.trim();
    this.task = null;
    this.parameters = [];
  }
  _initFromLine() {
    let line = this.line;
    // remove the first word
    let index = line.indexOf(" ");
    if (index === -1) {
      return false;
    }
    this.type = line.substring(0, index);
    // check if type is of the supported types
    if (!SupportedTypes.includes(this.type)) {
      return false;
    }
    // remove the first word
    line = line.substring(index + 1);
    this.task = line;
    // define a regexp for parameter start with < and end with >
    const regexp = /<([^>]*)>/g;

    let match = regexp.exec(line);
    while (match != null) {
      // add the parameter to the list
      this.parameters.push(match[1]);
      // get the next match
      match = regexp.exec(line);
    }
    return true;
  }
}
class Scenario {
  constructor(lines, index = 0) {
    this.lines = lines;
    this.index = index;
    this.scenarioInitiated = false;
    this.scenarioName = null;
    this.steps = [];
    this.endIndex = -1;
  }
  _initFromLines() {
    let i = this.index;
    for (; i < this.lines.length; i++) {
      let line = this.lines[i];
      if (line) {
        // trim the line and convert to lower case
        line = line.trim();
        if (!line.toLowerCase().startsWith("scenario:")) {
          continue;
        }
        this.scenarioInitiated = true;
        this.scenarioName = line.substring("scenario:".length).trim();
        i++;
        break;
      }
    }
    if (!this.scenarioInitiated) {
      return false;
    }
    for (; i < this.lines.length; i++) {
      let line = this.lines[i];
      if (!line) {
        continue;
      }

      if (line) {
        // trim the line and convert to lower case
        line = line.trim();
        if (line.toLowerCase().startsWith("scenario:")) {
          this.endIndex = i;
          return true;
        }
      }

      let step = new Step(line);
      if (!step._initFromLine()) {
        continue;
      }
      this.steps.push(step);
    }
    this.endIndex = i;
    return true;
  }
  _saveToFile(scenarioPath) {
    fs.mkdirSync(scenarioPath, { recursive: true });
    fs.writeFileSync(scenarioPath + "/scenario.json", JSON.stringify(this, null, 2));
  }
}

if (process.argv.length !== 5) {
  console.log("Usage: node cucumber.js <projectDir> <feature_file> <page_name>");
  process.exit(1);
}
let projectDir = process.argv[2];
let featureFile = process.argv[3];
let pageName = process.argv[4];

const aiAgent = new AIAgent(projectDir);
const context = {};
await aiAgent.initProjectAndEnvironment(context);
// let line =
//   '    When I create a new issue with title "<title>" and body "<body>"';
// let step = new Step(line);
// step._initFromLine();
// console.log(step);

// check if feature file exists
// create the path using projectDir and featureFile
const fullFeatureFilePath = path.join(projectDir, featureFile);
if (!fs.existsSync(fullFeatureFilePath)) {
  logger.error("Feature file not found: " + featureFile);
  process.exit(1);
}
const featureFileContent = fs.readFileSync(fullFeatureFilePath, "utf8");
const lines = featureFileContent.split("\n");

let scenario = new Scenario(lines);
if (!scenario._initFromLines()) {
  logger.error("Failed to parse feature file: " + featureFile);
  process.exit(1);
}

let first = true;
let previousTask = null;
let scenarioId = findNextIdInFolder("./tasks");
let scenarioPath = "./tasks" + "/" + scenarioId;
scenario._saveToFile(scenarioPath);
for (let i = 0; i < scenario.steps.length; i++) {
  let step = scenario.steps[i];
  let page = aiAgent.project.getPage(pageName);
  if (!page) {
    aiAgent.project.createNewPage(pageName, "/", false, "home page", false);
  }
  let taskProgress = new TaskProgress(-1, false, scenarioPath);

  await aiAgent.createNewStep(context, pageName, step.task, {}, step.type, taskProgress, false, first, previousTask);
  first = false;
  console.log(step);
  previousTask = step.task;
}
process.exit(0);
