import fs from "fs";
import { fileURLToPath } from "url";
import { dirname } from "path";
import path from "path";
import logger from "../../logger.js";

import { Node } from "./accessibility_node.js";
import { compareDumps } from "./accessibility_dump_compare.js";

const __filename = fileURLToPath(import.meta.url);
const __dirname = dirname(__filename);
const dompath = fs.readFileSync(path.resolve(__dirname, "./browser_side/locator.js"), "utf8");
const script = dompath.toString();

const cleanTree = (node) => {
  let foundEmptyNodes = false;
  if (node.children) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      if (child.children && child.children.length > 0) {
        let result = cleanTree(child);
        if (result) {
          foundEmptyNodes = true;
        }
      } else {
        if (child.axNode.ignored) {
          node.children.splice(i, 1);
          i--;
          foundEmptyNodes = true;
          continue;
        }
        if (child.roleType === "role" && child.role === "generic" && !child.name) {
          node.children.splice(i, 1);
          i--;
          foundEmptyNodes = true;
          continue;
        }
      }
    }
  }
  return foundEmptyNodes;
};
const collapseTree = (node) => {
  let collapse = false;
  if (node.children) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      if (child.ignored || (child.roleType === "role" && child.role === "generic" && !child.name)) {
        node.children.splice(i, 1);
        i--;
        if (child.children && child.children.length > 0) {
          for (let j = 0; j < child.children.length; j++) {
            const grandChild = child.children[j];
            node.children.splice(i + 1 + j, 0, grandChild);
          }
          i += child.children.length;
        }
        collapse = true;
        continue;
      }
      let result = collapseTree(child);
      if (result) {
        collapse = true;
      }
    }
  }
  return collapse;
};
const cleanActionStaticText = (node) => {
  if (
    (node.role === "link" ||
      node.role === "button" ||
      node.role === "tab" ||
      node.role === "menuitem" ||
      node.role === "menuitemcheckbox" ||
      node.role === "menuitemradio") &&
    node.children &&
    node.children.length > 0
  ) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      if (child.role === "StaticText" || child.role === "img") {
        node.children.splice(i, 1);
        i--;
      }
    }
  } else if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      cleanActionStaticText(child);
    }
  }
};
const unifyStaticText = (node) => {
  if (node.children && node.children.length > 0) {
    let text = "";
    let firstTextNode = null;
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      if (child.role === "StaticText") {
        if (!firstTextNode) {
          firstTextNode = child;
        } else {
          node.children.splice(i, 1);
          i--;
        }
        text += " " + child.name;
        firstTextNode.name = text.trim();
      } else {
        if (firstTextNode) {
          firstTextNode = null;
          text = "";
        }
        unifyStaticText(child);
      }
    }
  }
};

const collectTopLayerInformation = async (cdp, page, nodes) => {
  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i];
    if (!node.axNode.backendDOMNodeId) {
      logger.error("No backendDOMNodeId for node: " + node.toJSONString());
      continue;
    }
    const domNode = await cdp.send("DOM.resolveNode", {
      backendNodeId: node.axNode.backendDOMNodeId,
    });
    await page.evaluate(script);
    const res = await cdp.send("Runtime.callFunctionOn", {
      functionDeclaration: "function() { return document.getElementInfo(this); }",
      objectId: domNode.object.objectId,
      arguments: [
        {
          objectId: domNode.object.objectId,
        },
      ],
    });
    try {
      const resObj = JSON.parse(res.result.value);
      node.background = resObj.background;
      node.locator = resObj.locator;
      node.href = resObj.href;
    } catch (err) {
      logger.error("Failed to parse result: " + JSON.stringify(res));
    }
  }
};
const writeFile = (object, fileName, debugFiles) => {
  if (!debugFiles) {
    return;
  }
  fs.writeFileSync(
    fileName,
    JSON.stringify(
      object,
      (key, value) => {
        if (key !== "axNode" && key !== "nameType") {
          return value;
        }
        return undefined;
      },
      2
    )
  );
};
const buildAccessibilityTree = (node, nodesMap) => {
  const childrenIds = node.axNode.childIds;
  for (let i = 0; i < childrenIds.length; i++) {
    const id = childrenIds[i];
    if (nodesMap[id]) {
      const child = nodesMap[id];
      if (child.axNode.ignored) {
        if (child.axNode.childIds.length === 0) {
          continue;
        }
      }
      // if (!root.children) {
      //   root.children = [];
      // }
      if (!node.children) {
        node.children = [];
      }
      // if (child.axNode.ignored) {
      //   buildAccessibilityTree(root, child, nodesMap);
      // } else {
      node.children.push(child);
      buildAccessibilityTree(child, nodesMap);
      // }
    } else {
      logger.error("No node for id: " + id);
    }
  }
};
const flattenAllNodes = (root, nodesList) => {
  nodesList.push(root);
  for (let i = 0; i < root.children.length; i++) {
    const child = root.children[i];
    flattenAllNodes(child, nodesList);
  }
};
const removeBackgroundNodes = (node) => {
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      if (child.background) {
        node.children.splice(i, 1);
        i--;
      } else {
        removeBackgroundNodes(child);
      }
    }
  }
};
const removeTextLeafs = (node) => {
  if (node.children && node.children.length === 1 && node.children[0].role === "StaticText") {
    if (node.children[0].name === node.name || node.name === "") {
      node.name = node.children[0].name;
      node.children = [];
    }
  } else if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      removeTextLeafs(child);
    }
  }
};
const textboxValue = (node) => {
  if (
    node.role &&
    node.role === "textbox" &&
    node.children &&
    node.children.length === 1 &&
    node.children[0].role === "StaticText"
  ) {
    node.properties.value = node.children[0].name;
    node.children = [];
  } else if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      textboxValue(child);
    }
  }
};
const textboxBySilable = (node, silible) => {
  if (node.role && node.role === "textbox" && silible && silible.role === "StaticText") {
    node.nameAlternative = silible.name;
  } else if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      let childSilible = null;
      if (i > 0) {
        childSilible = node.children[i - 1];
      }
      textboxBySilable(child, childSilible);
    }
  }
};

const initParentProperty = (node) => {
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      child.parentId = node.nodeId;
      initParentProperty(child);
    }
  }
};
const generateUniqueRoleNameSet = (node, roleNameSet) => {
  if (node.role && node.name) {
    const key = node.role + "_" + node.name.trim();
    if (!roleNameSet[key]) {
      roleNameSet[key] = 0;
    }
    roleNameSet[key]++;
  }
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      generateUniqueRoleNameSet(child, roleNameSet);
    }
  }
};
const updateUniqueRoleName = (node, roleNameSet) => {
  if (node.role && node.name) {
    const key = node.role + "_" + node.name.trim();
    node.uniqueRoleName = roleNameSet[key] === 1;
  }
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      updateUniqueRoleName(child, roleNameSet);
    }
  }
};
const findUniqueRoleNameParent = (node, nodesMap) => {
  let parent = nodesMap[node.parentId];
  while (parent) {
    if (parent.uniqueRoleName) {
      return parent;
    }
    parent = nodesMap[parent.parentId];
  }
  return null;
};
const updateUniqueRoleNameParent = (node, nodesMap) => {
  if (!node.uniqueRoleName && node.name) {
    const parent = findUniqueRoleNameParent(node, nodesMap);
    if (parent) {
      const roleNameSet = {};
      generateUniqueRoleNameSet(parent, roleNameSet);
      if (roleNameSet[node.role + "_" + node.name.trim()] === 1) {
        node.uniqueParentRole = parent.role;
        node.uniqueParentName = parent.name;
      }
    }
  }
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      const child = node.children[i];
      updateUniqueRoleNameParent(child, nodesMap);
    }
  }
};
const collectAllRoleNodesPairs = (node, roleNodesPairs) => {
  if (node.role && node.name) {
    roleNodesPairs.push({ role: node.role, name: node.name });
  }
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      collectAllRoleNodesPairs(node.children[i], roleNodesPairs);
    }
  }
};
const updatePatternUnique = (node, roleNodesPairs) => {
  if (!node.pattern) {
    return;
  }
  for (let i = 0; i < roleNodesPairs.length; i++) {
    const pair = roleNodesPairs[i];
    if (pair.role === node.role && node.pattern === pair.pattern) {
      node.isPatternUnique = false;
      return;
    }
  }
  node.isPatternUnique = true;
};
const updatePatternUniqueForAll = (node, roleNodesPairs) => {
  updatePatternUnique(node, roleNodesPairs);
  if (node.children && node.children.length > 0) {
    for (let i = 0; i < node.children.length; i++) {
      updatePatternUniqueForAll(node.children[i], roleNodesPairs);
    }
  }
};

class AccessibilityDump {
  constructor() {
    this.type = "dump_screen";
  }

  async execute(context) {
    const playContext = context.playContext;
    const page = context.page;
    let nodes = [];
    let output = null;
    let cdp = null;
    let root = null;
    for (let i = 0; i < 3; i++) {
      nodes = [];
      try {
        cdp = await playContext.newCDPSession(page);
        await cdp.send("Page.enable", {});
        await cdp.send("Accessibility.enable", {});
        await cdp.send("DOM.enable", {});
        await cdp.send("Runtime.enable", {});
        await cdp.send("CSS.enable", {});
        const axNodes = await cdp.send("Accessibility.getFullAXTree", {});
        //const axNodeMap = {};
        const nodesMap = {};

        for (let i = 0; i < axNodes.nodes.length; i++) {
          const axNode = axNodes.nodes[i];
          const node = new Node(axNode);
          nodesMap[parseInt(axNode.nodeId)] = node;
        }

        root = new Node(axNodes.nodes[0]);
        buildAccessibilityTree(root, nodesMap);
        const roleNameSet = {};
        generateUniqueRoleNameSet(root, roleNameSet);
        updateUniqueRoleName(root, roleNameSet);
        const debugFiles = false;
        writeFile(root, "./1.json", debugFiles);
        // eslint-disable-next-line
        while (cleanTree(root)) {}
        writeFile(root, "./2.json", debugFiles);
        // eslint-disable-next-line
        while (collapseTree(root)) {}
        writeFile(root, "./3.json", debugFiles);
        cleanActionStaticText(root);
        writeFile(root, "./4.json", debugFiles);
        unifyStaticText(root);
        writeFile(root, "./5.json", debugFiles);
        const allNodes = [];
        flattenAllNodes(root, allNodes);
        await collectTopLayerInformation(cdp, page, allNodes);
        writeFile(root, "./6.json", debugFiles);
        removeBackgroundNodes(root);
        writeFile(root, "./7.json", debugFiles);
        textboxValue(root);
        textboxBySilable(root, null);
        removeTextLeafs(root);
        initParentProperty(root);
        updateUniqueRoleNameParent(root, nodesMap);
        const roleNodesPairs = [];
        collectAllRoleNodesPairs(root, roleNodesPairs);
        updatePatternUniqueForAll(root, roleNodesPairs);
        const idObject = { index: 0 };
        output = root.toJSONString(idObject, nodes, "\t");
        fs.writeFileSync("./out.json", output);
        break;
      } catch (err) {
        logger.error("Failed to dump page: " + err);
        if (i == 2) {
          throw err;
        }
      } finally {
        await cdp.detach();
      }
    }

    let pageTitle = root.name;
    const url = await page.url();
    context.url = url;
    //context.interactionElementsText = output;
    context.elements = nodes;
    context.pageText = compareDumps(output, context.previousPageText);
    context.previousPageText = output;
    // generate page dump
    try {
      var s = pageTitle;
      if (!s) {
        // generate random string
        s = Math.random().toString(36).substring(7);
      }
      let filename = s.replaceAll(/[^a-z0-9]/gi, "_").toLowerCase() + ".json";
      if (!fs.existsSync("./pages")) {
        fs.mkdir("./pages", { recursive: true });
      }
      fs.writeFileSync("./pages/" + filename, output, "utf-8");
    } catch (err) {
      console.error(err);
    }
  }
}

export { AccessibilityDump };
