import fs from "node:fs";
import Walker from "node-source-walk";
import path from "node:path";
import vm from "vm";
import logger from "../../../logger.js";

class CodePage {
  constructor(sourceFileName = null) {
    this.sourceFileName = sourceFileName;
    this.cleanFileName = path.basename(sourceFileName).split(".")[0];
    this._init();
  }
  _init() {
    this.nextCodePartIndex = 0;
    this.fileContent = null;
    this.imports = [];
    this.pageClass = null;
    this.classProperties = [];
    this.variables = [];
    this.expressions = [];
    this.orderedCodeParts = [];
  }
  save() {
    if (this.sourceFileName !== null) {
      fs.writeFileSync(this.sourceFileName, this.fileContent, "utf8");
      return true;
    } else {
      logger.error("sourceFileName is null");
      return false;
    }
  }

  generateModel = (fileContent = null) => {
    if (fileContent !== null) {
      this.fileContent = fileContent;
    } else {
      this.fileContent = fs.readFileSync(this.sourceFileName, "utf8");
    }
    const walker = new Walker();
    let rootNode = null;
    walker.walk(this.fileContent, (node) => {
      if (
        node.constructor.name === "SourceLocation" ||
        node.constructor.name === "Position" ||
        node.constructor.name === "Object"
      ) {
        return;
      }

      if (node.type === "ImportDeclaration") {
        this.addCodePart(node);
        if (rootNode === null) {
          rootNode = node;
        }
      } else if (node.type === "ExpressionStatement") {
        if (node.parent.filter((n) => n === rootNode).length > 0) {
          this.addCodePart(node);
        }
      } else if (node.type === "ClassDeclaration") {
        this.addCodePart(node);
      } else if (node.type === "ClassProperty") {
        this.addCodePart(node);
      } else if (node.type === "VariableDeclaration") {
        if (node.parent.filter((n) => n === rootNode).length > 0) {
          this.addCodePart(node);
        }
      } else {
        //console.log("node", node);
      }

      // if (false) {
      //   console.log(
      //     "node " + node.loc.start.line + "," + node.loc.start.column,
      //     node
      //   );
      // }
    });
  };
  addCucumberStep(type, cucumberLine, method, parameters) {
    let code = "";
    code += `${type}(${JSON.stringify(cucumberLine)}, async (${parameters.join(", ")}) => {\n`;
    code += `  let _page = await ${this.getClassName()}.init();\n`;
    code += `  await _page.${method}(${parameters.join(", ")});\n`;
    code += "});\n";
    let injectIndex = this._getCucumberInjectionIndex();
    this._insertCode(code, injectIndex);
  }
  addInfraCommand(methodName, description, stepParameters, stepCodeLines) {
    let code = "";
    code += "  /**\n";
    code += `   * ${description}\n`;
    //const keys = Object.keys(stepParameters);
    stepParameters.forEach((key) => {
      code += `   * @param {string} ${key} ${key.replaceAll("_", " ")}\n`;
    });
    code += "   */\n";
    code += `  ${methodName} = async (${stepParameters.join(", ")}) => {\n`;
    code += "    this.navigate();\n";
    stepCodeLines.forEach((line) => {
      code += `    ${line}\n`;
    });
    code += "\n  };\n";

    let injectIndex = this._getMethodInjectionIndex();
    this._insertCode(code, injectIndex);
  }
  _insertCode(code, injectIndex) {
    let newFileContent = this.fileContent.substring(0, injectIndex) + code + this.fileContent.substring(injectIndex);
    //fs.writeFileSync(this.sourceFileName, newFileContent, "utf8");

    // update the model
    this._init();
    try {
      this.generateModel(newFileContent);
    } catch (e) {
      logger.error("new code", newFileContent);
      throw e;
    }
  }
  // _locatorToCode(el) {
  //   if (el.role) {
  //     let name = el.role[1].name;
  //     if (typeof name === "string") {
  //       name = JSON.stringify(name);
  //     }
  //     return `    { role: ["${el.role[0]}", { name: ${name} }] },\n`;
  //   } else if (el.css) {
  //     return `    { css: "${el.css}" },\n`;
  //   } else if (el.text) {
  //     return `    { text: ${JSON.stringify(el.text)} },\n`;
  //   } else {
  //     throw new Error("Unknown element type" + JSON.stringify(el));
  //   }
  // }
  _elementToCode(key, element) {
    let code = "";
    code += `  ${key}:${JSON.stringify(element)}`;
    // if (Array.isArray(element)) {
    //   code += "  [\n";
    //   element.forEach((el) => {
    //     code += this._locatorToCode(el);
    //   });
    //   code += "  ]";
    // } else if (typeof element === "object") {
    //   code += this._locatorToCode(element);
    // } else {
    //   throw new Error("Unknown element type" + JSON.stringify(element));
    // }
    return code;
  }

  insertElements(elements) {
    let code = "const elements = {\n";
    const keys = Object.keys(elements);
    const lines = [];
    keys.forEach((key) => {
      lines.push(this._elementToCode(key, elements[key]));
    });
    code += lines.join(",\n");
    code += "\n};\n";
    let startEnd = this._getVariableStartEnd("elements");
    this._replaceCode(code, startEnd[0], startEnd[1]);
  }
  _replaceCode(code, injectIndexStart, injectIndexEnd) {
    let newFileContent =
      this.fileContent.substring(0, injectIndexStart) + code + this.fileContent.substring(injectIndexEnd);
    this._init();
    console.log("new code", newFileContent);
    this.generateModel(newFileContent);
  }
  _getMethodInjectionIndex() {
    return this.pageClass.end - 1;
  }
  _getCucumberInjectionIndex() {
    return this.fileContent.length;
  }

  _getVariableStartEnd(variableName) {
    const codePart = this.variables.filter((cp) => cp.node.declarations[0].id.name === variableName)[0];
    if (codePart === undefined) {
      return null;
    }
    return [codePart.node.start, codePart.node.end];
  }
  getVariableDeclarationCode(variableName) {
    const codePart = this.variables.filter((cp) => cp.node.declarations[0].id.name === variableName)[0];
    if (codePart === undefined) {
      return null;
    }
    return codePart.codePart;
  }
  getVariableDeclarationAsObject(variableName) {
    const code = this.getVariableDeclarationCode(variableName);
    if (code === null) {
      return null;
    }
    let value = null;
    let vmContext = {
      console: console,
      value: value,
    };
    vm.createContext(vmContext);
    vm.runInContext(code + "\nvalue = " + variableName + ";", vmContext);
    //console.log("value", vmContext.value);
    return vmContext.value;
  }
  isLoginRequired() {
    const code = this.getVariableDeclarationCode("login_required");
    if (code === null) {
      return true;
    }
    let value = null;
    let vmContext = {
      console: console,
      value: value,
    };
    vm.createContext(vmContext);
    vm.runInContext(code + "\nvalue = login_required;", vmContext);
    return vmContext.value;
  }
  isNavigateOnInit() {
    const code = this.getVariableDeclarationCode("navigate_on_init");
    if (code === null) {
      return true;
    }
    let value = null;
    let vmContext = {
      console: console,
      value: value,
    };
    vm.createContext(vmContext);
    vm.runInContext(code + "\nvalue = navigate_on_init;", vmContext);
    return vmContext.value;
  }
  getName() {
    const code = this.getVariableDeclarationCode("name");
    if (code === null) {
      return null;
    }
    let value = null;
    let vmContext = {
      console: console,
      value: value,
    };
    vm.createContext(vmContext);
    vm.runInContext(code + "\nvalue = name;", vmContext);
    return vmContext.value;
  }

  addCodePart(node) {
    const codePart = new CodePart(node, this.fileContent, this.nextCodePartIndex);
    this.nextCodePartIndex++;
    if (node.type === "ImportDeclaration") {
      this.imports.push(codePart);
    } else if (node.type === "ExpressionStatement") {
      this.expressions.push(codePart);
    } else if (node.type === "ClassDeclaration") {
      this.pageClass = codePart;
    } else if (node.type === "ClassProperty") {
      this.classProperties.push(codePart);
    } else if (node.type === "VariableDeclaration") {
      this.variables.push(codePart);
    }
  }

  // getMethodCode(methodName) {
  //   const codePart = this.getMethodCodePart(methodName);
  //   return codePart.codePart;
  // }
  getMethodsNames() {
    const methods = [];
    this.classProperties.forEach((codePart) => {
      if (!codePart.node.key.name.startsWith("_")) {
        methods.push(codePart.node.key.name);
      }
    });
    return methods;
  }
  getMethodParametersValues(methodName, parameters) {
    const params = this.getMethodParameters(methodName);
    const values = [];
    for (let i = 0; i < params.length; i++) {
      const param = params[i];
      values.push(parameters[param]);
    }
    return values;
  }

  getClassName() {
    return this.pageClass.node.id.name;
  }
  // getClassComments() {
  //   const comments = this.pageClass.node.leadingComments;
  //   if (comments === undefined || comments.length === 0) {
  //     return null;
  //   }
  //   const commentsValues = comments.map((comment) => comment.value);
  //   return commentsValues.join("\n");
  // }
  getMethodParameters(methodName) {
    const codePart = this.classProperties.find((cp) => cp.node.key.name === methodName);

    if (!codePart) {
      // Handle the case where methodName is not found
      return [];
    }

    const params = codePart.node.value.params;
    return params.map((param) => param.name);
  }

  // getVarableValue(variableName) {
  //   const codePart = this.variables.filter(
  //     (cp) => cp.node.declarations[0].id.name === variableName
  //   )[0];
  //   if (codePart === undefined) {
  //     return null;
  //   }
  //   return codePart.node.declarations[0].init.value;
  // }

  // getPageSignature(withComments = false) {
  //   let signature = "page file name: " + this.cleanFileName + "\n";

  //   if (withComments) {
  //     const comment = this.getClassComments();
  //     if (comment) {
  //       signature += "/*" + comment + "*/\n";
  //     }
  //   }
  //   signature += "class " + this.getClassName() + " ...\n";
  //   this.classProperties.forEach((codePart) => {
  //     let methodName = codePart.node.key.name;
  //     signature += this._getMethodSignature(methodName, withComments) + "\n";
  //   });
  //   return signature + "\n";
  // }

  getCodeWithoutImports(removeCucumber = true) {
    let text = this.fileContent;
    if (removeCucumber) {
      const expressionsLoc = [];
      this.expressions.forEach((codePart) => {
        expressionsLoc.push([codePart.start, codePart.end]);
      });
      text = cutSubString(text, expressionsLoc);
    }
    let lastImportEnd = 0;
    this.imports.forEach((codePart) => {
      lastImportEnd = codePart.end;
    });

    text = text.substring(lastImportEnd);
    return text;
  }
  hasMethod(methodName) {
    return this.classProperties.filter((cp) => cp.node.key.name === methodName).length > 0;
  }
  _getMethodComments(methodName) {
    const codePart = this.classProperties.filter((cp) => cp.node.key.name === methodName)[0];
    if (codePart === undefined) {
      return null;
    }
    const comments = codePart.node.leadingComments;
    if (comments === undefined || comments.length === 0) {
      return null;
    }
    const commentsValues = comments.map((comment) => comment.value);
    return commentsValues.join("\n");
  }
  /**
   * get the description and parameters of a method
   * @param {string} methodName
   * @returns a dictionary with name, description and params
   */
  getMethodCommentsData(methodName) {
    const codePart = this.classProperties.filter((cp) => cp.node.key.name === methodName)[0];
    if (codePart === undefined) {
      return null;
    }
    const comments = codePart.node.leadingComments;
    if (comments === undefined || comments.length === 0) {
      return null;
    }
    const commentsValues = comments.map((comment) => comment.value);
    const fullCommentText = commentsValues.join("\n");
    // split trim and remove * if exist from the beginning of each line
    const commentLines = fullCommentText.split("\n").map((line) => line.trim().replace(/^\*/, "").trim());
    const descriptions = [];
    for (let i = 0; i < commentLines.length; i++) {
      const line = commentLines[i];
      if (!line) {
        continue;
      }
      if (line.startsWith("@")) {
        break;
      }
      descriptions.push(line);
    }
    // join the descriptions into one line seperated by '.' and replace 2 dots with 1
    const description = descriptions.join(".").replace("..", ".");
    // extract parameters from the comment
    const params = [];
    for (let i = 0; i < commentLines.length; i++) {
      const line = commentLines[i];
      if (line.startsWith("@param")) {
        params.push(this._processParametersLine(line));
      }
    }
    return { name: methodName, description, params };
  }

  /**
   * get the method in a format that can be used in the gpt prompt
   * @param {string} methodName
   * @returns a string formated to be used in the commands gpt prompt
   */
  getMethodInCommandFormat(methodName) {
    const methodComment = this.getMethodCommentsData(methodName);
    if (methodComment === null) {
      return null;
    }
    return `${methodName}: ${methodComment.description}, args: ${methodComment.params
      .map((p) => `"${p.name}": "<${p.description.replaceAll(" ", "_")}>"`)
      .join(", ")}`;
  }

  _processParametersLine(paramLine) {
    // split by multiple spaces
    const paramParts = paramLine.split(/\s+/);
    // remove @param
    paramParts.shift();
    // remove type
    let type = null;
    if (paramParts.length > 0 && paramParts[0].startsWith("{") && paramParts[0].endsWith("}")) {
      type = paramParts.shift().replace("{", "").replace("}", "");
    }
    let name = null;
    if (paramParts.length > 0) {
      name = paramParts.shift();
    }
    let description = null;
    if (paramParts.length > 0) {
      description = paramParts.join(" ");
    }
    return { type, name, description };
  }
  getMethodCodePart(methodName) {
    const codePart = this.classProperties.filter((cp) => cp.node.key.name === methodName)[0];
    if (codePart === undefined) {
      return null;
    }
    return codePart;
  }

  // _getMethodSignature(methodName, withComments = false) {
  //   let signature = "";

  //   const codePart = this.getMethodCodePart(methodName);
  //   if (withComments) {
  //     const comment = this._getMethodComments(methodName);
  //     if (comment) {
  //       signature += "/*" + comment + "*/\n";
  //     }
  //   }
  //   signature +=
  //     "async function " +
  //     methodName +
  //     "(" +
  //     this.getMethodParameters(methodName).join(", ") +
  //     ") ...";
  //   return signature;
  // }
}

class CodePart {
  constructor(node, fileContent, index, codePart = null) {
    this.node = node;
    this.fileContent = fileContent;
    if (this.node !== null) {
      this.start = node.loc.start.index;
      this.end = node.loc.end.index;
      this.codePart = fileContent.substring(this.start, this.end);
    } else {
      this.pagePart = codePart;
      this.start = -1;
      this.end = -1;
    }
    this.index = index;
  }
}

const cutSubString = (str, substrings) => {
  let result = str;
  let offset = 0;
  substrings.forEach((substring) => {
    const start = substring[0];
    const end = substring[1];
    result = result.substring(0, start - offset) + result.substring(end - offset);
    offset += end - start;
  });
  return result;
};

export { CodePage };
