import { Configuration, OpenAIApi } from "openai";
import fs from "fs";
import logger from "../../../logger.js";
import { PromptProxy } from "prompt-eng";

const config = JSON.parse(fs.readFileSync("./config.json", "utf8"));
const configuration = new Configuration({
  apiKey: config.openaiKey,
});
process.env.NODE_ENV = "dev";
const openai = new PromptProxy(new OpenAIApi(configuration));
const executeChatRequest = async (messages, model = "gpt-4", temperature = 0.0) => {
  const completion = await openai
    .createChatCompletion({
      model: model,
      messages: messages,
      temperature: temperature,
    })
    .then((response) => {
      return response;
    })
    .catch((error) => {
      if (error.response) {
        logger.error(error.response.data);
        logger.error(error.response.status);
      }
      throw error;
      //logger.error(error.response);
    });
  return completion;
};
const executeRequest = async (prompt, model = "text-davinci-003", temperature = 0.0) => {
  const response = await openai
    .createCompletion({
      model,
      prompt,
      temperature,
      max_tokens: 256,
      top_p: 1,
      frequency_penalty: 0,
      presence_penalty: 0,
    })
    .then((response) => {
      return response;
    })
    .catch((error) => {
      if (error.response) {
        logger.error(error.response.data);
        logger.error(error.response.status);
      }
      throw error;
      //logger.error(error.response);
    });
  return response;
};
const generatePrompt = (promptFile, parameters, chat = true) => {
  let promptFileContent = fs.readFileSync(promptFile, "utf8");
  for (const key in parameters) {
    const value = parameters[key];
    promptFileContent = promptFileContent.replaceAll(`{${key}}`, value);
  }
  if (!chat) return { promptFileContent };
  const messages = [];
  const parts = promptFileContent.split("####");
  for (let i = 0; i < parts.length; i++) {
    if (!parts[i].trim()) continue;
    const role = parts[i];
    const prompt = parts[i + 1];
    messages.push({ role, content: prompt.trim() });
    //    console.error("####prompt####", `role ${role}ֿֿ`);
    //    console.error(prompt.trim());
    i++;
  }
  return { messages, promptFileContent };
};

const runPrompt = async (
  promptFileName,
  parameters,
  model = "gpt-4",
  temperature = 0.0,
  chat = true,
  retry = 3,
  timeout = 60000
) => {
  if (chat) {
    const messages = generatePrompt(promptFileName, parameters);
    // const cache = {
    //   prompt: messages.promptFileContent,
    //   parameters,
    //   model,
    // };
    logger.debug("##############################################");
    logger.debug(messages.messages);
    let completion = await retryAsyncFunctionWithTimeout(
      async function () {
        return await executeChatRequest(messages.messages, model, temperature);
      },
      timeout,
      retry
    );

    // return result
    if (completion.data.choices[0]) {
      return completion.data.choices[0];
    }
    throw new Error("No response from OpenAI");
  } else {
    const messages = generatePrompt(promptFileName, parameters, false);
    // const cache = {
    //   prompt: messages.promptFileContent,
    //   parameters,
    //   model,
    // };
    logger.debug("##############################################");
    logger.debug(messages.messages);

    let completion = await retryAsyncFunctionWithTimeout(
      async function () {
        return await executeRequest(messages.messages, model, temperature);
      },
      timeout,
      retry
    );
    if (completion.data.choices[0]) {
      return completion.data.choices[0];
    }
    throw new Error("No response from OpenAI");
  }
};
function retryAsyncFunctionWithTimeout(asyncFunction, timeout, maxRetries) {
  return new Promise((resolve, reject) => {
    let retries = 0;

    function executeFunction() {
      const timer = setTimeout(() => {
        retries++;
        if (retries <= maxRetries) {
          console.log(`Retrying (${retries}/${maxRetries})...`);
          executeFunction();
        } else {
          clearTimeout(timer);
          reject(new Error(`Timeout exceeded after ${timeout} ms`));
        }
      }, timeout);

      asyncFunction()
        .then((result) => {
          clearTimeout(timer);
          resolve(result);
        })
        .catch((error) => {
          clearTimeout(timer);
          retries++;
          if (retries <= maxRetries) {
            console.log(`Retrying (${retries}/${maxRetries})...`);
            executeFunction();
          } else {
            reject(error);
          }
        });
    }

    executeFunction();
  });
}

export { executeChatRequest, runPrompt, generatePrompt };
