import fs from "fs";
import { changeToTaskLogger } from "../../logger.js";
import logger from "../../logger.js";

const TasksTypes = {
  CREATE_STEP: "CREATE_STEP",
};
const findNextIdInFolder = (folder) => {
  if (!fs.existsSync(folder)) {
    fs.mkdirSync(folder, { recursive: true });
    return 1;
  }
  const directoryContents = fs.readdirSync(folder);
  const digitDirectories = directoryContents.filter((entry) => {
    const entryPath = `${folder}/${entry}`;
    const isDirectory = fs.statSync(entryPath).isDirectory();
    const containsOnlyDigits = /^\d+$/.test(entry);
    return isDirectory && containsOnlyDigits;
  });
  if (digitDirectories.length === 0) {
    return 1;
  }
  return Math.max(...digitDirectories.map((folder) => parseInt(folder))) + 1;
};
class TaskProgress {
  constructor(id = -1, readExisting = false, tasksPath = "./tasks") {
    this.type = null;
    this.id = id;
    this.task = null;
    this.status = null;
    this.start_time = null;
    this.end_time = null;
    this.memory = [];
    this.commands = [];
    this.lastCommand = null;
    this.logFile = null;
    this.success = null;
    this.error = null;
    this.tasksPath = tasksPath;
    this.taskPath = null;
    this.steps = [];
    this.progress = "Start";
    this.taskParameters = {};
    this.shadow = false;
    this.screenshotId = 0;
    if (id === -1) {
      id = findNextIdInFolder(this.tasksPath);
      this.id = id;
    }
    this.taskPath = this.tasksPath + "/" + this.id;
    if (readExisting) {
      this.readFromFile();
    }
  }

  getScreenShotPath() {
    this.screenshotId++;
    if (!fs.existsSync(this.taskPath + "/screenshots")) {
      fs.mkdirSync(this.taskPath + "/screenshots", { recursive: true });
    }
    return this.taskPath + "/screenshots/" + this.screenshotId + ".png";
  }
  isActive() {
    return !this.shadow && this.id >= 0;
  }
  readFromFile() {
    const taskFilePath = this.taskPath + "/task.json";
    if (!fs.existsSync(taskFilePath)) {
      return;
    }
    const taskFile = fs.readFileSync(taskFilePath);
    const task = JSON.parse(taskFile);
    Object.assign(this, task);
  }
  setLoggerFile() {
    if (this.shadow) {
      return;
    }
    let logFile = this.taskPath + "/out.log";
    fs.mkdirSync(this.taskPath, { recursive: true });
    changeToTaskLogger(logFile);
    this.logFile = logFile;
  }

  updateLastCommand(properties) {
    // clone the properties to lastCommand
    if (!this.lastCommand) {
      return;
    }
    Object.assign(this.lastCommand, properties);
    this._saveToFile();
  }
  _saveToFile() {
    if (this.shadow) {
      return;
    }
    if (!this.id) {
      return;
    }
    const taskFilePath = this.taskPath + "/task.json";
    fs.mkdirSync(this.taskPath, { recursive: true });
    fs.writeFileSync(taskFilePath, JSON.stringify(this, null, 2));
  }
  startTask(type, task, progress, taskParameters = {}) {
    if (this.shadow) {
      return;
    }
    this.setTaskType(type);
    this.taskParameters = taskParameters;
    this.start_time = new Date().toISOString();
    this.task = task;
    this.progress = progress;
    this._saveToFile();
    logger.info(`Starting. ID: ${this.id} of type ${type}:\n${JSON.stringify(taskParameters, null, 2)})}`);
  }
  updateProgress(progress = null) {
    if (this.shadow) {
      return;
    }
    if (progress) {
      this.progress = progress;
    }
    this._saveToFile();
  }
  setError(error) {
    if (this.shadow) {
      return;
    }
    this.error = error;
    this.staus = "error";
    this._saveToFile();
  }
  setDone(progress = "Done", success = true) {
    if (this.shadow) {
      return;
    }
    this.success = success;
    if (success) {
      this.status = "success";
    } else {
      this.status = "failed";
    }

    this.end_time = new Date().toISOString();
    this._saveToFile();
    this.updateProgress(progress);
  }
  setTaskType(type) {
    this.type = type;
    this.status = "running";
    switch (type) {
      case TasksTypes.CREATE_STEP:
        this.steps = ["Start", "Login/Navigate", "Record steps", "Generate code", "Done"];
        this.name = "Create new step";
        break;
      default:
        throw new Error(`Unknown task type ${type}`);
    }
  }
  addProgressCommand() {
    this.lastCommand = {};
    this.commands.push(this.lastCommand);
    this._saveToFile();
  }
}
export { TaskProgress, TasksTypes, findNextIdInFolder };
